/* -*-c-*-
 *
 * $Id: f-prime.c,v 1.9 2004/04/01 21:28:41 mdw Exp $
 *
 * Prime fields with Montgomery arithmetic
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: f-prime.c,v $
 * Revision 1.9  2004/04/01 21:28:41  mdw
 * Normal basis support (translates to poly basis internally).  Rewrite
 * EC and prime group table generators in awk, so that they can reuse data
 * for repeated constants.
 *
 * Revision 1.8  2004/04/01 12:50:09  mdw
 * Add cyclic group abstraction, with test code.  Separate off exponentation
 * functions for better static linking.  Fix a buttload of bugs on the way.
 * Generally ensure that negative exponents do inversion correctly.  Add
 * table of standard prime-field subgroups.  (Binary field subgroups are
 * currently unimplemented but easy to add if anyone ever finds a good one.)
 *
 * Revision 1.7  2004/03/27 17:54:11  mdw
 * Standard curves and curve checking.
 *
 * Revision 1.6  2004/03/23 15:19:32  mdw
 * Test elliptic curves more thoroughly.
 *
 * Revision 1.5  2004/03/23 12:08:26  mdw
 * Random field-element selection.
 *
 * Revision 1.4  2004/03/21 22:52:06  mdw
 * Merge and close elliptic curve branch.
 *
 * Revision 1.3.4.3  2004/03/21 22:39:46  mdw
 * Elliptic curves on binary fields work.
 *
 * Revision 1.3.4.2  2004/03/20 00:13:31  mdw
 * Projective coordinates for prime curves
 *
 * Revision 1.3.4.1  2003/06/10 13:43:53  mdw
 * Simple (non-projective) curves over prime fields now seem to work.
 *
 * Revision 1.3  2003/05/15 23:25:59  mdw
 * Make elliptic curve stuff build.
 *
 * Revision 1.2  2002/01/13 13:48:44  mdw
 * Further progress.
 *
 * Revision 1.1  2001/04/29 18:12:33  mdw
 * Prototype version.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/sub.h>

#include "field.h"
#include "mpmont.h"
#include "mprand.h"

/*----- Main code ---------------------------------------------------------*/

typedef struct fctx {
  field f;
  mpmont mm;
} fctx;

/* --- Field operations --- */

static void fdestroy(field *ff)
  { fctx *f = (fctx *)ff; mpmont_destroy(&f->mm); DESTROY(f); }

static mp *frand(field *ff, mp *d, grand *r)
  { fctx *f = (fctx *)ff; return (mprand_range(d, f->mm.m, r, 0)); }

static mp *fin(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff;
  mp_div(0, &d, x, f->mm.m);
  return (mpmont_mul(&f->mm, d, d, f->mm.r2));
}

static mp *fout(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; return (mpmont_reduce(&f->mm, d, x)); }

static int fzerop(field *ff, mp *x) { return (!MP_LEN(x)); }

static mp *fneg(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; return (mp_sub(d, f->mm.m, x)); }

static mp *fadd(field *ff, mp *d, mp *x, mp *y) {
  fctx *f = (fctx *)ff; d = mp_add(d, x, y);
  if (d->f & MP_NEG) d = mp_add(d, d, f->mm.m);
  else if (MP_CMP(d, >, f->mm.m)) d = mp_sub(d, d, f->mm.m);
  return (d);
}

static mp *fsub(field *ff, mp *d, mp *x, mp *y) {
  fctx *f = (fctx *)ff; d = mp_sub(d, x, y);
  if (d->f & MP_NEG) d = mp_add(d, d, f->mm.m);
  else if (MP_CMP(d, >, f->mm.m)) d = mp_sub(d, d, f->mm.m);
  return (d);
}

static mp *fmul(field *ff, mp *d, mp *x, mp *y)
  { fctx *f = (fctx *)ff; return (mpmont_mul(&f->mm, d, x, y)); }

static mp *fsqr(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; d = mp_sqr(d, x);
  return (mpmont_reduce(&f->mm, d, d));
}

static mp *finv(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; d = mpmont_reduce(&f->mm, d, x);
  mp_gcd(0, 0, &d, f->mm.m, d); return (mpmont_mul(&f->mm, d, d, f->mm.r2));
}

static mp *freduce(field *ff, mp *d, mp *x)
  { fctx *f = (fctx *)ff; mp_div(0, &d, x, f->mm.m); return (d); }

static mp *fsqrt(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; d = mpmont_reduce(&f->mm, d, x);
  d = mp_modsqrt(d, d, f->mm.m); if (!d) return (d);
  return (mpmont_mul(&f->mm, d, d, f->mm.r2));
}

static mp *fdbl(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; d = mp_lsl(d, x, 1);
  if (MP_CMP(d, >, f->mm.m)) d = mp_sub(d, d, f->mm.m);
  return (d);
}

static mp *ftpl(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; MP_DEST(d, MP_LEN(x) + 1, x->f);
  MPX_UMULN(d->v, d->vl, x->v, x->vl, 3);
  while (MP_CMP(d, >, f->mm.m)) d = mp_sub(d, d, f->mm.m);
  return (d);
}

static mp *fqdl(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff; d = mp_lsl(d, x, 2);
  while (MP_CMP(d, >, f->mm.m)) d = mp_sub(d, d, f->mm.m);
  return (d);
}

static mp *fhlv(field *ff, mp *d, mp *x) {
  fctx *f = (fctx *)ff;
  if (!MP_LEN(x)) { MP_COPY(x); MP_DROP(d); return (x); }
  if (x->v[0] & 1) { d = mp_add(d, x, f->mm.m); x = d; }
  return (mp_lsr(d, x, 1));
}

/* --- Field operations table --- */

static field_ops fops = {
  FTY_PRIME, "prime",
  fdestroy, frand, field_stdsamep,
  fin, fout,
  fzerop, fneg, fadd, fsub, fmul, fsqr, finv, freduce, fsqrt,
  0,
  fdbl, ftpl, fqdl, fhlv
};

/* --- @field_prime@ --- *
 *
 * Arguments:	@mp *p@ = the characteristic of the field
 *
 * Returns:	A pointer to the field.
 *
 * Use:		Creates a field structure for a prime field of size %$p$%,
 *		using Montgomery reduction for arithmetic.
 */

field *field_prime(mp *p)
{
  fctx *f = CREATE(fctx);
  f->f.ops = &fops;
  mpmont_create(&f->mm, p);
  f->f.zero = MP_ZERO;
  f->f.one = f->mm.r;
  f->f.m = f->mm.m;
  f->f.nbits = mp_bits(p);
  f->f.noctets = (f->f.nbits + 7) >> 3;
  return (&f->f);
}

/*----- That's all, folks -------------------------------------------------*/
